@@ -1,5 +1,5 @@
 /* Copyright (C) 1995, 1996, 1997, 1998, 2000, 2001, 2004, 2006, 2010,
- *   2012, 2013 Free Software Foundation, Inc.
+ *   2012, 2013, 2014 Free Software Foundation, Inc.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License
@@ -40,9 +40,7 @@
 #ifdef HAVE_STRING_H
 #include <string.h>
 #endif
-#ifdef HAVE_UNISTD_H
 #include <unistd.h>
-#endif
 
 /* For Windows... */
 #ifdef HAVE_IO_H
@@ -80,7 +78,8 @@ SCM_DEFINE (scm_error_scm, "scm-error", 5, 0, 0,
 	    "@code{system-error} then it should be a list containing the\n"
 	    "Unix @code{errno} value; If @var{key} is @code{signal} then it\n"
 	    "should be a list containing the Unix signal number; If\n"
-	    "@var{key} is @code{out-of-range} or @code{wrong-type-arg},\n"
+	    "@var{key} is @code{out-of-range}, @code{wrong-type-arg},\n"
+	    "or @code{keyword-argument-error}, "
             "it is a list containing the bad value; otherwise\n"
 	    "it will usually be @code{#f}.")
 #define FUNC_NAME s_scm_error_scm
