@@ -1,5 +1,6 @@
-/*	Copyright (C) 1995,1996,1997, 2000, 2001, 2003, 2004, 2006, 2008, 2009, 2010, 2011 Free Software Foundation, Inc.
- * 
+/* Copyright (C) 1995, 1996, 1997, 2000, 2001, 2003, 2004, 2006, 2008,
+ *   2009, 2010, 2011, 2012, 2014 Free Software Foundation, Inc.
+ *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License
  * as published by the Free Software Foundation; either version 3 of
@@ -26,6 +27,7 @@
 #include <wchar.h>
 #endif
 
+#include <math.h>
 #include <unistr.h>
 
 #include "libguile/_scm.h"
@@ -43,6 +45,7 @@
 extern double floor();
 #endif
 
+#define SCM_MIN(A, B) ((A) < (B) ? (A) : (B))
 
 unsigned long 
 scm_string_hash (const unsigned char *str, size_t len)
@@ -159,18 +162,16 @@ scm_hasher(SCM obj, unsigned long n, size_t d)
     if (SCM_CHARP(obj))
       return (unsigned)(scm_c_downcase(SCM_CHAR(obj))) % n;
     switch (SCM_UNPACK (obj)) {
-#ifndef SICP
-    case SCM_UNPACK(SCM_EOL):
+    case SCM_EOL_BITS:
       d = 256; 
       break;
-#endif
-    case SCM_UNPACK(SCM_BOOL_T):
+    case SCM_BOOL_T_BITS:
       d = 257; 
       break;
-    case SCM_UNPACK(SCM_BOOL_F):
+    case SCM_BOOL_F_BITS:
       d = 258; 
       break;
-    case SCM_UNPACK(SCM_EOF_VAL):
+    case SCM_EOF_VAL_BITS:
       d = 259; 
       break;
     default: 
@@ -192,7 +193,7 @@ scm_hasher(SCM obj, unsigned long n, size_t d)
       case scm_tc16_real:
 	{
 	  double r = SCM_REAL_VALUE (obj);
-	  if (floor (r) == r) 
+	  if (floor (r) == r && !isinf (r) && !isnan (r))
 	    {
 	      obj = scm_inexact_to_exact (obj);
 	      return scm_to_ulong (scm_modulo (obj, scm_from_ulong (n)));
@@ -224,33 +225,38 @@ scm_hasher(SCM obj, unsigned long n, size_t d)
 	significant_bits = (scm_t_uintptr) SCM_POINTER_VALUE (obj) >> 4UL;
 	return (size_t) significant_bits  % n;
       }
+    case scm_tcs_struct:
+      return scm_i_struct_hash (obj, n, d);
     case scm_tc7_wvect:
     case scm_tc7_vector:
-      {
-	size_t len = SCM_SIMPLE_VECTOR_LENGTH (obj);
-	if (len > 5)
-	  {
-	    size_t i = d/2;
-	    unsigned long h = 1;
-	    while (i--)
-	      {
-		SCM elt = SCM_SIMPLE_VECTOR_REF (obj, h % len);
-		h = ((h << 8) + (scm_hasher (elt, n, 2))) % n;
-	      }
-	    return h;
-	  }
-	else
-	  {
-	    size_t i = len;
-	    unsigned long h = (n)-1;
-	    while (i--)
-	      {
-		SCM elt = SCM_SIMPLE_VECTOR_REF (obj, h % len);
-		h = ((h << 8) + (scm_hasher (elt, n, d/len))) % n;
-	      }
-	    return h;
-	  }
-      }
+      if (d > 0)
+        {
+          size_t len, i, d2;
+          unsigned long h;
+
+          len = SCM_SIMPLE_VECTOR_LENGTH (obj);
+          if (len > 5)
+            {
+              i = d / 2;
+              h = 1;
+              d2 = SCM_MIN (2, d - 1);
+            }
+          else
+            {
+              i = len;
+              h = n - 1;
+              d2 = len > 0 ? (d - 1) / len : 0;
+            }
+
+          while (i--)
+            {
+              SCM elt = SCM_SIMPLE_VECTOR_REF (obj, h % len);
+              h = ((h << 8) + (scm_hasher (elt, n, d2))) % n;
+            }
+          return h;
+        }
+      else
+        return 1;
     case scm_tcs_cons_imcar: 
     case scm_tcs_cons_nimcar:
       if (d) return (scm_hasher (SCM_CAR (obj), n, d/2)
